#pragma once

#include <stdint.h>
#include <stdlib.h>

/* Max value for any buffer, to not squash the stack. */
#define BUFFER_MAX 4096

#ifndef __GNUC__
#define inline
#endif

/* Macros */

#define BRACKETS_THAT(that) (that)

#ifndef MAX
#define MAX(a, b) ((a) < (b) ? (b) : (a))
#endif /* MAX */

#ifndef MIN
#define MIN(a, b) ((a) > (b) ? (b) : (a))
#endif /* MIN */

#define RETURN_IF(cond, msg, ret)                                       \
    if (cond) {                                                         \
        LOG_ERROR("%s", msg);                                           \
        return ret;                                                     \
    }
#define GOTO_IF(cond, msg, label)                                       \
    if (cond) {                                                         \
        LOG_ERROR("%s", msg);                                           \
        goto label;                                                     \
    }

#define GOTO_UNLESS(cond, msg, label)   GOTO_IF(!(cond), msg, label)
#define RETURN_UNLESS(cond, msg, ret)   RETURN_IF(!(cond), msg, ret)
#define NOTHING                         /* Usefull to return nothing. */

#define STRTOL(ret, str, endptr, err_flag)                              \
{                                                                       \
    err_flag = 0;                                                       \
    errno    = 0;                                                       \
    ret      = str == NULL ? 0 : strtol(str, &(endptr), 0);             \
    err_flag = errno != 0 || endptr == str;                             \
}

#define ERROR   1
#define WARN    2
#define INFO    3
#define DEBUG   4
#define LOG(format, level, section, ...)                                            \
    fprintf(stderr, " %c %s%s: " format "\n",                                       \
            level == ERROR ? '!' : level == WARN ? '*' : level == INFO ? '.' : ' ', \
            sizeof(section) > sizeof("") ? ("[" section "] ") : "",                 \
            __func__,                                                               \
            __VA_ARGS__)
#define LOG_INFO_SCT(section, format, ...)      LOG(format, INFO, section, __VA_ARGS__)
#define LOG_WARN_SCT(section, format, ...)      LOG(format, WARN, section, __VA_ARGS__)
#define LOG_ERROR_SCT(section, format, ...)     LOG(format, ERROR, section, __VA_ARGS__)
#define LOG_DEBUG_SCT(section, format, ...)     LOG(format, DEBUG, section, __VA_ARGS__)
#define LOG_INFO(format, ...)                   LOG_INFO_SCT("", format, __VA_ARGS__)
#define LOG_WARN(format, ...)                   LOG_WARN_SCT("", format, __VA_ARGS__)
#define LOG_ERROR(format, ...)                  LOG_DEBUG_SCT("", format, __VA_ARGS__)
#define LOG_DEBUG(format, ...)                  LOG_DEBUG_SCT("", format, __VA_ARGS__)

/* Defines for lektor */

#define INI_MAX_LINE_LEN    512
#define INI_MAX_SECTION_LEN 80

#define URL_MAX_LEN         1024
#define DEFAULT_URL         "https://kurisu.iiens.net"
#define GET_ID_JSON         DEFAULT_URL "/api_karas.php?id=%ld"
#define GET_ID_FILE         DEFAULT_URL "/download.php?id=%ld"

#define SELF_EXECUTABLE_LINUX           "/proc/self/exe"
#define SELF_EXECUTABLE_FREEBSD         "/proc/curproc/file"
#define SELF_EXECUTABLE_SOLARIS         "/proc/self/path/a.out"


#define LKT_MAX_SQLITE_STATEMENT        1024
#define PROTECTED_DATABASE              "disk"

#define LKT_DATABASE_NAME_KID           "id"
#define LKT_DATABASE_NAME_KNAME         "source_name"
#define LKT_DATABASE_NAME_KTITLE        "song_title"
#define LKT_DATABASE_NAME_KCAT          "category"
#define LKT_DATABASE_NAME_KTYPE         "song_type"
#define LKT_DATABASE_NAME_KAUTHOR       "author_name"
#define LKT_DATABASE_NAME_KAUTHOR_YEAR  "author_year"
#define LKT_DATABASE_NAME_KLANG         "language"
#define LKT_DATABASE_KARA_COLUMNT_ANY   "any_col"
#define LKT_DATABASE_KARA_ALL           "string"

#define LEKTOR_TAG_MAX                  256
#define LKT_MESSAGE_ARGS_MAX            32
#define LKT_MESSAGE_MAX                 2048
#define LKT_DEFAULT_LIST_SIZE           10

typedef volatile enum {
    MPD_IDLE_NONE               = 0,

    MPD_IDLE_DATABASE           = ( 1 << 1  ),
    MPD_IDLE_UPDATE             = ( 1 << 2  ),
    MPD_IDLE_STORED_PLAYLIST    = ( 1 << 3  ),
    MPD_IDLE_PLAYLIST           = ( 1 << 4  ),
    MPD_IDLE_PLAYER             = ( 1 << 5  ),
    MPD_IDLE_MIXER              = ( 1 << 6  ),
    MPD_IDLE_OUTPUT             = ( 1 << 7  ),
    MPD_IDLE_OPTIONS            = ( 1 << 8  ),
    MPD_IDLE_PARTITION          = ( 1 << 9  ),
    MPD_IDLE_STICKER            = ( 1 << 10 ),
    MPD_IDLE_SUBSCRIPTION       = ( 1 << 11 ),
    MPD_IDLE_MESSAGE            = ( 1 << 12 ),

    MPD_IDLE_ALL =
        MPD_IDLE_DATABASE | MPD_IDLE_UPDATE | MPD_IDLE_STORED_PLAYLIST  |
        MPD_IDLE_PLAYLIST | MPD_IDLE_PLAYER | MPD_IDLE_MIXER            |
        MPD_IDLE_OUTPUT | MPD_IDLE_OPTIONS | MPD_IDLE_PARTITION         |
        MPD_IDLE_STICKER | MPD_IDLE_SUBSCRIPTION | MPD_IDLE_MESSAGE,
} mpd_idle_flag;

#define LKT_BACKLOG         32
#define COMMAND_LIST_MAX    64
#define BUFFER_OUT_MAX      16
#define MPD_VERSION         "0.21.16"

/* Macros */

#define SQLITE_PREPARE(db, stmt, SQL, goto_label)                               \
    if (sqlite3_prepare_v2((sqlite3 *) db, SQL, -1, &(stmt), 0) != SQLITE_OK) { \
        LOG_ERROR_SCT("DB", "Failed to prepare statement: %s",                  \
                      sqlite3_errmsg((sqlite3 *) db));                          \
        goto goto_label;                                                        \
    }

#define SQLITE_EXEC(db, SQL, goto_label)                                        \
    if (sqlite3_exec((sqlite3 *) db, SQL, NULL, NULL, NULL) != SQLITE_OK) {     \
        LOG_ERROR_SCT("DB", "Failed to exec statement: %s",                     \
                      sqlite3_errmsg((sqlite3 *) db));                          \
        goto goto_label;                                                        \
    }

#define SQLITE_BIND_TEXT(db, stmt, pos, text, error)                    \
    if (sqlite3_bind_text(stmt, pos, text, -1, 0) != SQLITE_OK) {       \
        LOG_ERROR_SCT("DB", "Failed to bind text %s at pos %d: %s",     \
                      text, pos, sqlite3_errmsg((sqlite3 *) db));       \
        goto error;                                                     \
    }

#define SQLITE_BIND_INT(db, stmt, pos, integer, error)                  \
    if (sqlite3_bind_int(stmt, pos, integer) != SQLITE_OK) {            \
        LOG_ERROR_SCT("DB", "Failed to bind int %d at pos %d: %s",      \
                      integer, pos, sqlite3_errmsg((sqlite3 *) db));    \
        goto error;                                                     \
    }

#define SQLITE_STEP(db, stmt, code, error)                              \
    if (sqlite3_step(stmt) != code) {                                   \
        LOG_ERROR_SCT("DB", "Failed to step: %s",                       \
                      sqlite3_errmsg((sqlite3 *) db));                  \
        goto error;                                                     \
    }

#define SQLITE_STEP_ROW(db, stmt, error)    SQLITE_STEP(db, stmt, SQLITE_ROW, error)
#define SQLITE_STEP_DONE(db, stmt, error)   SQLITE_STEP(db, stmt, SQLITE_DONE, error)
#define SQLITE_STEP_OK(db, stmt, error)     SQLITE_STEP(db, stmt, SQLITE_OK, error)

#define SQLITE_DO_ROLLBACK(db)                                          \
    sqlite3_exec((sqlite3 *) db, "ROLLBACK TRANSACTION;\n", NULL, NULL, NULL);

#define STR_MATCH(str1, str2)           (! strcasecmp(str1, str2))
#define STR_NMATCH(str1, str2, n)       (! strncasecmp(str1, str2, n))
